#include "application.h"
#include "thread.h"
#include "utils.h"
#include "r3dutils.h"

#include "R3DSDK.h"

using namespace R3DSDK;

vector<Thread*> threads;

unsigned int runningThreadsAmount(){
    for (unsigned int i=0; i <threads.size(); i++) {
        if (threads.at(i)->isFinished()){
            threads.at(i)->call();
            threads.at(i)->~Thread();
            threads.erase(threads.begin()+i);
            i--;
        }
    }

    return threads.size();
}

Application::Rects* Application::getOrientationRects(string orientation, int width, int height){
    Application::Rects* rects=nullptr;
    if(!orientation.empty()){
        rects= new Application::Rects();
        if (!orientation.compare(Settings::getInstance()->VERTICALLY_STRING)){
            rects->left = Rect(0, 0, width/2, height);
            rects->right = Rect (width/2, 0, width/2, height);
        }else if(!orientation.compare(Settings::getInstance()->HORIZONTALLY_STRING)){
            rects->left = Rect(0, 0, width, height/2);
            rects->right = Rect(0, height/2, width, height/2);
        }
    }

    return rects;
}

int Application::executeNative(string input[], string orientation, DisparityCheckerInterface* disparityChecker, unsigned int limit, string output, string sequence){

    InitializeSdk(".", OPTION_RED_NONE);

    Clip* captureLeft = new Clip(input[0].c_str());
    Clip* captureRight = new Clip(input[1].c_str());

    int width = captureLeft->Width();
    int height = captureLeft->Height();

    Application::Rects* rects = this->getOrientationRects(orientation, width, height);

    size_t memoryNeeded = width * height * 3U * 2U;
    unsigned int adjustedLeft = memoryNeeded;                       unsigned int adjustedRight = memoryNeeded;

    unsigned char* bufferLeft = R3DUtils::allocate(adjustedLeft);   unsigned char* bufferRight = R3DUtils::allocate(adjustedRight);

    VideoDecodeJob leftJob;                                         VideoDecodeJob rightJob;
    leftJob.BytesPerRow = width * 2U;                               rightJob.BytesPerRow = width * 2U;
    leftJob.OutputBufferSize = memoryNeeded;                        rightJob.OutputBufferSize = memoryNeeded;
    leftJob.Mode = DECODE_FULL_RES_PREMIUM;                         rightJob.Mode = DECODE_FULL_RES_PREMIUM;
    leftJob.OutputBuffer = bufferLeft;                              rightJob.OutputBuffer = bufferRight;
    leftJob.PixelType = PixelType_16Bit_RGB_Planar;                 rightJob.PixelType = PixelType_16Bit_RGB_Planar;

    int size = int(captureLeft->VideoFrameCount());

    Mat frameLeft;
    Mat frameRight;

    vector<int> sequences = Utils::getSequences(sequence);

    for(int id=0; id < size && captureLeft->DecodeVideoFrame(id, leftJob) == DSDecodeOK && captureRight->DecodeVideoFrame(id, rightJob) == DSDecodeOK; id++){
        frameLeft = R3DUtils::toMat(bufferLeft, width, height);
        frameRight = R3DUtils::toMat(bufferRight, width, height);

        if(rects != nullptr){
            frameLeft = frameLeft(rects->left);
            frameRight = frameLeft(rects->right);
        }

        if (find(sequences.begin(), sequences.end(), id+1) != sequences.end() || sequences.empty()){
            if(R3DUtils::isHorizontalFlip(input[0])) flip(frameLeft, frameLeft, 1);
            if(R3DUtils::isVerticalFlip(input[0])) flip(frameLeft, frameLeft, 0);

            if(R3DUtils::isHorizontalFlip(input[1])) flip(frameRight, frameRight, 1);
            if(R3DUtils::isVerticalFlip(input[1])) flip(frameRight, frameRight, 0);

            Thread* thread = new Thread(this, disparityChecker, frameLeft, frameRight, id+1, size, Utils::getFilename(input[0]).c_str(), output);
            threads.push_back(thread);
            thread->start();
        }

        while(runningThreadsAmount()>=limit) QThread::msleep(THREAD_WAIT);
    }

    free(bufferLeft - adjustedLeft);
    free(bufferRight - adjustedRight);
    delete captureLeft, captureRight;
    FinalizeSdk();

    while(runningThreadsAmount()>0) QThread::msleep(THREAD_WAIT);

    return 0;
}

int Application::executeMovie(string input[], string orientation, DisparityCheckerInterface* disparityChecker, unsigned int limit, string output, string sequence){

    VideoCapture captureLeft(input[0].c_str());
    VideoCapture captureRight(input[1].c_str());

    int width = int(captureLeft.get(CV_CAP_PROP_FRAME_WIDTH));
    int height = int(captureLeft.get(CV_CAP_PROP_FRAME_HEIGHT));

    Application::Rects* rects = this->getOrientationRects(orientation, width, height);

    int size = int(captureLeft.get(CV_CAP_PROP_FRAME_COUNT));

    Mat frameLeft;
    Mat frameRight;

    vector<int> sequences = Utils::getSequences(sequence);

    for(int id=0; id < size && captureLeft.read(frameLeft) && captureRight.read(frameRight); id++){
        if(rects != nullptr){
            frameLeft = frameLeft(rects->left);
            frameRight = frameRight(rects->right);
        }

        if (find(sequences.begin(), sequences.end(), id+1) != sequences.end() || sequences.empty()){
            Thread* thread = new Thread(this, disparityChecker, frameLeft, frameRight, id+1, size, Utils::getFilename(input[0]), output);
            threads.push_back(thread);
            thread->start();
        }

        while(runningThreadsAmount()>=limit) QThread::msleep(THREAD_WAIT);
    }

    while(runningThreadsAmount()>0) QThread::msleep(THREAD_WAIT);

    return 0;
}

int Application::executeImage(string input[], string orientation, DisparityCheckerInterface* disparityChecker, string output){

    Mat frameLeft = imread(input[0].c_str(), CV_LOAD_IMAGE_ANYCOLOR);
    Mat frameRight= imread(input[1].c_str(), CV_LOAD_IMAGE_ANYCOLOR);

    int width = int(frameLeft.cols);
    int height = int(frameLeft.rows);

    Application::Rects* rects = this->getOrientationRects(orientation, width, height);

    if(rects != nullptr){
        frameLeft = frameLeft(rects->left);
        frameRight = frameRight(rects->right);
    }

    Thread* thread = new Thread(this, disparityChecker, frameLeft, frameRight, 1, 1, Utils::getFilename(input[0]), output);
    threads.push_back(thread);
    thread->start();

    while(runningThreadsAmount()>0) QThread::msleep(THREAD_WAIT);

    return 0;
}

int Application::execute(string input[], string orientation, string disparity, string threads, string output, string parameters, string sequence){

    DisparityCheckerInterface* disparityChecker = DisparityCheckerInterface::getInstance(disparity, parameters);

    switch(Utils::getFormat(input)){
        case Utils::Format::NATIVE :{
            return this->executeNative(input, orientation, disparityChecker, atoi(threads.c_str()), output, sequence);
            break;
        }
        case Utils::Format::MOVIE :{
            return this->executeMovie(input, orientation, disparityChecker, atoi(threads.c_str()), output, sequence);
            break;
        }
        case Utils::Format::IMAGE :{
            return this->executeImage(input, orientation, disparityChecker, output);
            break;
        }
        case Utils::Format::UNKNOWN :{
            Logger::error("ConsoleApplication::run - [extension]", "Unsupported file extension or incorrect format");
            return -1;
            break;
        }
    }

    return 0;
}
