#include "correlation.h"

#include <time.h>
#include <iostream>

float minDistance(vector<DMatch> matches){
    float min = FLT_MAX;
    for(DMatch match: matches)
        if(match.distance < min) min = match.distance;

    return min;
}

Correlation* Correlation::execute(DisparityCheckerInterface* disparityChecker, Mat left, Mat right){
    int horizontalDisparity = (left.cols * atoi(disparityChecker->getParameter("WIDTH").c_str()))/100;
    int verticalDisparity = (left.rows * atoi(disparityChecker->getParameter("HEIGHT").c_str()))/100;


    int margin = 28;

    double start = clock();
    vector<KeyPoint> globalKeypointsLeft, globalKeypointsRight;
    vector<DMatch> globalMatches;
    for(int i=margin+horizontalDisparity; i+margin+horizontalDisparity < left.cols && i+margin+horizontalDisparity < right.cols; i+=horizontalDisparity){
        for(int j=margin+verticalDisparity; j+margin+verticalDisparity < left.rows && j+margin+verticalDisparity < right.rows; j+=verticalDisparity){
            Mat windowLeft = left(Rect(i-(margin+horizontalDisparity), j-(margin+verticalDisparity), 2*(margin+horizontalDisparity), 2*(margin+verticalDisparity)));
            Mat windowRight = right(Rect(i-(margin+horizontalDisparity), j-(margin+verticalDisparity), 2*(margin+horizontalDisparity), 2*(margin+verticalDisparity)));

            vector<KeyPoint> keypointsLeft, keypointsRight;
            Ptr<FeatureDetector> detector = FeatureDetector::create(disparityChecker->getParameter("DETECTOR"));
            detector->detect(windowLeft, keypointsLeft);
            detector->detect(windowRight, keypointsRight);

            Mat descriptorsLeft, descriptorsRight;
            Ptr<DescriptorExtractor> extractor = DescriptorExtractor::create(disparityChecker->getParameter("EXTRACTOR"));
            extractor->compute( windowLeft, keypointsLeft, descriptorsLeft);
            extractor->compute( windowRight, keypointsRight, descriptorsRight);

            if (descriptorsLeft.cols != descriptorsRight.cols) continue;

            vector<DMatch> matchesLeft, matchesRight, goodMatches;
            Ptr<DescriptorMatcher> matcher = DescriptorMatcher::create(disparityChecker->getParameter("MATCHER"));
            matcher->match(descriptorsLeft, descriptorsRight, matchesLeft);
            matcher->match(descriptorsRight, descriptorsLeft, matchesRight);

            for (unsigned int k = 0; k < matchesLeft.size(); k++) {

                DMatch matchLeft = matchesLeft.at(k);
                DMatch matchRight = matchesRight.at(matchLeft.trainIdx);

                if (matchRight.trainIdx == matchLeft.queryIdx && matchRight.distance<=minDistance(matchesRight) &&  matchRight.distance <= extractor->descriptorSize()/2 && matchLeft.distance<=minDistance(matchesLeft) && matchLeft.distance <= extractor->descriptorSize()/2) {
                    goodMatches.push_back(matchLeft);

                }
            }

            for(unsigned int k=0; k < keypointsLeft.size(); k++){
                keypointsLeft.at(k).pt.x +=(i-(margin+horizontalDisparity));
                keypointsLeft.at(k).pt.y +=(j-(margin+verticalDisparity));
            }

            for(unsigned int k=0; k < keypointsRight.size(); k++){
                keypointsRight.at(k).pt.x +=(i-(margin+horizontalDisparity));
                keypointsRight.at(k).pt.y +=(j-(margin+verticalDisparity));
            }

            for(unsigned int k=0; k < goodMatches.size(); k++){
                goodMatches.at(k).queryIdx += globalKeypointsLeft.size();
                goodMatches.at(k).trainIdx += globalKeypointsRight.size();
                globalMatches.push_back(goodMatches.at(k));
            }

            globalKeypointsLeft.insert(globalKeypointsLeft.end(), keypointsLeft.begin(), keypointsLeft.end());
            globalKeypointsRight.insert(globalKeypointsRight.end(), keypointsRight.begin(), keypointsRight.end());
        }
    }
    double stop = clock();
    double time = 1000*(stop-start)/CLOCKS_PER_SEC;

    Correlation* correlation = new Correlation(globalKeypointsLeft, globalKeypointsRight, globalMatches);
    return correlation;
}

Correlation::Correlation(vector<KeyPoint> keypointsLeft, vector<KeyPoint> keypointsRight, vector<DMatch> matches)
{
    this->keypointsLeft = keypointsLeft;
    this->keypointsRight = keypointsRight;
    this->matches = matches;

}

vector<KeyPoint> Correlation::getKeypointsLeft(){
    return this->keypointsLeft;
}
vector<KeyPoint> Correlation::getKeypointsRight(){
    return this->keypointsRight;
}
vector<DMatch> Correlation::getMatches(){
    return this->matches;
}
