#include "disparitychecker.h"
#include "verticaldisparitychecker.h"
#include "rotationdisparitychecker.h"
#include "scaledisparitychecker.h"

#include "settings.h"

#include <opencv2/opencv.hpp>

DisparityChecker::DisparityChecker(string parameters){
    this->parameters = parameters;
}

Mat DisparityChecker::correct(Mat image, double value){
    Mat correct; image.copyTo(correct);

    return correct;
}
string getPhase(string hierarchy, int index){
    int position = 0;
    string phases(hierarchy); phases+=Settings::getInstance()->SEMICOLON_STRING;
    string phase;

    int counter = 0;
    while ((position = phases.find(Settings::getInstance()->SEMICOLON_STRING)) != string::npos) {
        counter++;
        phase = phases.substr(0, position);
        if(counter == index) return phase;
        phases.erase(0, position+1);
    }
    return Settings::getInstance()->EMPTY_STRING;
}

Mat getImage(string phase, string name, vector<Result*> results){
    for(Result* result: results){
        if(!phase.compare(result->getName()) || phase.empty()){
            for(Result::KeyImage* image: result->getImages()){
                if(!name.compare(image->getKey())) return image->getImage();
            }
        }
    }
    Mat image;
    return image;
}

vector<Result*> DisparityChecker::execute(Correlation* correlation, Mat left, Mat right){

    if(left.type() == CV_16UC3 || right.type() == CV_16UC3){
        left.convertTo(left, CV_8U, 1.0/256.0);
        right.convertTo(right, CV_8U, 1.0/256.0);
    }

    string recurrency = this->getParameter("RECURRENCY");

    vector<Result*> results;
    string hierarchy = this->getParameter("HIERARCHY");
    for(int i=1; getPhase(hierarchy, i) != Settings::getInstance()->EMPTY_STRING; i++){
        string phase = getPhase(hierarchy, i);
        if (!phase.compare("VERTICAL")){
            DisparityCheckerInterface* verticalDisparityChecker = new VerticalDisparityChecker(this->parameters);
            vector<Result*> result = verticalDisparityChecker->execute(correlation, left, right);
            results.insert(results.end(), result.begin(), result.end());
            if(!recurrency.compare("TRUE")){
                left = getImage(Settings::getInstance()->EMPTY_STRING, "CORRECT_LEFT", result);
                right = getImage(Settings::getInstance()->EMPTY_STRING, "CORRECT_RIGHT", result);
            }
        }
        if (!phase.compare("ROTATION")){
            DisparityCheckerInterface* rotationDisparityChecker = new RotationDisparityChecker(this->parameters);
            vector<Result*> result = rotationDisparityChecker->execute(correlation, left, right);
            results.insert(results.end(), result.begin(), result.end());
            if(!recurrency.compare("TRUE")){
                left = getImage(Settings::getInstance()->EMPTY_STRING, "CORRECT_LEFT", result);
                right = getImage(Settings::getInstance()->EMPTY_STRING, "CORRECT_RIGHT", result);
            }
        }
        if (!phase.compare("SCALE")){
            DisparityCheckerInterface* scaleDisparityChecker = new ScaleDisparityChecker(this->parameters);
            vector<Result*> result = scaleDisparityChecker->execute(correlation, left, right);
            results.insert(results.end(), result.begin(), result.end());
            if(!recurrency.compare("TRUE")){
                left = getImage(Settings::getInstance()->EMPTY_STRING, "CORRECT_LEFT", result);
                right = getImage(Settings::getInstance()->EMPTY_STRING, "CORRECT_RIGHT", result);
            }
        }
    }

    return results;
}

Mat DisparityChecker::summary(vector<Result*> results){
    if(results.size()){
        Mat image = getImage(Settings::getInstance()->EMPTY_STRING, "CORRECT_LEFT", results);
        int width = image.cols;
        int height = image.rows;
        int type = image.type();

        Mat summary(3 * height, 3 * width, type, Scalar(240,240,240));

        Mat verticalCircles = getImage("VERTICAL", "CIRCLES", results);
        if(verticalCircles.rows && verticalCircles.cols){
            string text = "VERTICAL";
            int type = 8;
            Size size = getTextSize(text, FONT_HERSHEY_DUPLEX, 6, 6, &type);
            Point position = Point((1 * width)-(1.5*(size.width)), (1 * height)-(1.5*(size.height)));
            verticalCircles.copyTo(summary(Rect(0 * width, 0 * height, 1 * width, 1 * height)));
            putText(summary, text, position, FONT_HERSHEY_DUPLEX, 6, CV_RGB(255,255,255), 6, type, false);
        }

        Mat rotationCircles = getImage("ROTATION", "CIRCLES", results);
        if(rotationCircles.rows && rotationCircles.cols){
            string text = "ROTATION";
            int type = 8;
            Size size = getTextSize(text, FONT_HERSHEY_DUPLEX, 6, 6, &type);
            Point position = Point((2 * width)-(1.5*(size.width)), (1 * height)-(1.5*(size.height)));
            rotationCircles.copyTo(summary(Rect(1 * width, 0 * height, 1 * width, 1 * height)));
            putText(summary, text, position, FONT_HERSHEY_DUPLEX, 6, CV_RGB(255,255,255), 6, type, false);
        }

        Mat scaleCircles = getImage("SCALE", "CIRCLES", results);
        if(scaleCircles.rows && scaleCircles.cols){
            string text = "SCALE";
            int type = 8;
            Size size = getTextSize(text, FONT_HERSHEY_DUPLEX, 6, 6, &type);
            Point position = Point((3 * width)-(1.5*(size.width)), (1 * height)-(1.5*(size.height)));
            scaleCircles.copyTo(summary(Rect(2 * width, 0 * height, 1 * width, 1 * height)));
            putText(summary, text, position, FONT_HERSHEY_DUPLEX, 6, CV_RGB(255,255,255), 6, type, false);
        }

        Mat after = getImage("SCALE", "ANAGLYPH_AFTER", results);
        if(after.rows && after.cols){
            Mat scale;
            resize(after, scale, Size(2 * width, 2 * height));
            string text = "ANAGLYPH";
            int type = 8;
            Size size = getTextSize(text, FONT_HERSHEY_DUPLEX, 6, 6, &type);
            Point position = Point((2 * width)-(1.5*(size.width)), (3 * height)-(1.5*(size.height)));
            scale.copyTo(summary(Rect(0 * width, 1 * height, 2 * width, 2 * height)));
            putText(summary, text, position, FONT_HERSHEY_DUPLEX, 6, CV_RGB(255,255,255), 6, type, false);
            scale.release();
        }else{
            after = getImage("ROTATION", "ANAGLYPH_AFTER", results);
            if(after.rows && after.cols){
                Mat scale;
                resize(after, scale, Size(2 * width, 2 * height));
                string text = "ANAGLYPH";
                int type = 8;
                Size size = getTextSize(text, FONT_HERSHEY_DUPLEX, 6, 6, &type);
                Point position = Point((2 * width)-(1.5*(size.width)), (3 * height)-(1.5*(size.height)));
                scale.copyTo(summary(Rect(0 * width, 1 * height, 2 * width, 2 * height)));
                putText(summary, text, position, FONT_HERSHEY_DUPLEX, 6, CV_RGB(255,255,255), 6, type, false);
                scale.release();
            }else{
                after = getImage("VERTICAL", "ANAGLYPH_AFTER", results);
                if(after.rows && after.cols){
                    Mat scale;
                    resize(after, scale, Size(2 * width, 2 * height));
                    string text = "ANAGLYPH";
                    int type = 8;
                    Size size = getTextSize(text, FONT_HERSHEY_DUPLEX, 6, 6, &type);
                    Point position = Point((2 * width)-(1.5*(size.width)), (3 * height)-(1.5*(size.height)));
                    scale.copyTo(summary(Rect(0 * width, 1 * height, 2 * width, 2 * height)));
                    putText(summary, text, position, FONT_HERSHEY_DUPLEX, 6, CV_RGB(255,255,255), 6, type, false);
                    scale.release();
                }
            }
        }

        Mat correctLeft = getImage("SCALE", "CORRECT_LEFT", results);
        if(correctLeft.rows && correctLeft.cols){
            string text = "LEFT";
            int type = 8;
            Size size = getTextSize(text, FONT_HERSHEY_DUPLEX, 6, 6, &type);
            Point position = Point((3 * width)-(1.5*(size.width)), (2 * height)-(1.5*(size.height)));
            correctLeft.copyTo(summary(Rect(2 * width, 1 * height, 1 * width, 1 * height)));
            putText(summary, text, position, FONT_HERSHEY_DUPLEX, 6, CV_RGB(255,255,255), 6, type, false);
        }else{
            correctLeft = getImage("ROTATION", "CORRECT_LEFT", results);
            if(correctLeft.rows && correctLeft.cols){
                string text = "LEFT";
                int type = 8;
                Size size = getTextSize(text, FONT_HERSHEY_DUPLEX, 6, 6, &type);
                Point position = Point((3 * width)-(1.5*(size.width)), (2 * height)-(1.5*(size.height)));
                correctLeft.copyTo(summary(Rect(2 * width, 1 * height, 1 * width, 1 * height)));
                putText(summary, text, position, FONT_HERSHEY_DUPLEX, 6, CV_RGB(255,255,255), 6, type, false);
            }else{
                correctLeft = getImage("VERTICAL", "CORRECT_LEFT", results);
                if(correctLeft.rows && correctLeft.cols){
                    string text = "LEFT";
                    int type = 8;
                    Size size = getTextSize(text, FONT_HERSHEY_DUPLEX, 6, 6, &type);
                    Point position = Point((3 * width)-(1.5*(size.width)), (2 * height)-(1.5*(size.height)));
                    correctLeft.copyTo(summary(Rect(2 * width, 1 * height, 1 * width, 1 * height)));
                    putText(summary, text, position, FONT_HERSHEY_DUPLEX, 6, CV_RGB(255,255,255), 6, type, false);
                }
            }
        }

        Mat correctRight = getImage("SCALE", "CORRECT_RIGHT", results);
        if(correctRight.rows && correctRight.cols){
            string text = "RIGHT";
            int type = 8;
            Size size = getTextSize(text, FONT_HERSHEY_DUPLEX, 6, 6, &type);
            Point position = Point((3 * width)-(1.5*(size.width)), (3 * height)-(1.5*(size.height)));
            correctRight.copyTo(summary(Rect(2 * width, 2 * height, 1 * width, 1 * height)));
            putText(summary, text, position, FONT_HERSHEY_DUPLEX, 6, CV_RGB(255,255,255), 6, type, false);
        }else{
            correctRight = getImage("ROTATION", "CORRECT_RIGHT", results);
            if(correctRight.rows && correctRight.cols){
                string text = "RIGHT";
                int type = 8;
                Size size = getTextSize(text, FONT_HERSHEY_DUPLEX, 6, 6, &type);
                Point position = Point((3 * width)-(1.5*(size.width)), (3 * height)-(1.5*(size.height)));
                correctRight.copyTo(summary(Rect(2 * width, 2 * height, 1 * width, 1 * height)));
                putText(summary, text, position, FONT_HERSHEY_DUPLEX, 6, CV_RGB(255,255,255), 6, type, false);
            }else{
                correctRight = getImage("VERTICAL", "CORRECT_RIGHT", results);
                if(correctRight.rows && correctRight.cols){
                    string text = "RIGHT";
                    int type = 8;
                    Size size = getTextSize(text, FONT_HERSHEY_DUPLEX, 6, 6, &type);
                    Point position = Point((3 * width)-(1.5*(size.width)), (3 * height)-(1.5*(size.height)));
                    correctRight.copyTo(summary(Rect(2 * width, 2 * height, 1 * width, 1 * height)));
                    putText(summary, text, position, FONT_HERSHEY_DUPLEX, 6, CV_RGB(255,255,255), 6, type, false);
                }
            }
        }

        line(summary, Point(0 * width, 1 * height), Point(3 * width, 1 * height), Scalar(0, 0, 0), 15);
        line(summary, Point(2 * width, 2 * height), Point(3 * width, 2 * height), Scalar(0, 0, 0), 15);

        line(summary, Point(1 * width, 0 * height), Point(1 * width, 1 * height), Scalar(0, 0, 0), 15);
        line(summary, Point(2 * width, 0 * height), Point(2 * width, 3 * height), Scalar(0, 0, 0), 15);

        return summary;
    }else{
        Mat summary(600, 800, CV_8UC3, Scalar(255,255,255));
        string text = "Wrong application arguments ...";
        int type = 8;
        Size size = getTextSize(text, FONT_HERSHEY_DUPLEX, 1, 1, &type);
        Point position = Point((summary.cols/2)-(size.width/2), (summary.rows/2)+(size.height/2));
        putText(summary, text, position, FONT_HERSHEY_DUPLEX, 1, CV_RGB(0,0,0), 1, type, false);
        return summary;
    }
}
