#include "rotationdisparitychecker.h"
#include "disparity.h"
#include "statistic.h"
#include "visualize.h"

#define PI 3.14159265359

RotationDisparityChecker::RotationDisparityChecker(string parameters){
    this->parameters = parameters;
}

Mat RotationDisparityChecker::correct(Mat image, double value){
    Mat correct; image.copyTo(correct);

    Point2f center(image.cols/2.0F, image.rows/2.0F);
    Mat transformation = getRotationMatrix2D(center, value, 1);
    warpAffine(image, correct, transformation, image.size());

    return correct;
}

vector<Result*> RotationDisparityChecker::execute(Correlation* correlation, Mat left, Mat right){
    if(left.type() == CV_16UC3 || right.type() == CV_16UC3){
        left.convertTo(left, CV_8U, 1.0/256.0);
        right.convertTo(right, CV_8U, 1.0/256.0);
    }

    vector<Result*> results;

    if(correlation == nullptr) correlation = Correlation::execute(this, left, right);
    vector<KeyPoint> keypointsLeft = correlation->getKeypointsLeft();
    vector<KeyPoint> keypointsRight = correlation->getKeypointsRight();
    vector<DMatch> matches = correlation->getMatches();

    vector<Result::KeyValue*> data;
    vector<Result::KeyImage*> images;

    Mat before = Visualize::anaglyph(left, right);
    images.push_back(new Result::KeyImage(string("ANAGLYPH_BEFORE"), before));

    double start = clock();
    vector<Disparity*> disparities;
    for (unsigned int i = 0; i < matches.size(); i++) {
        DMatch match = matches.at(i);
        KeyPoint keypointLeft = keypointsLeft.at(match.queryIdx);
        KeyPoint keypointRight = keypointsRight.at(match.trainIdx);

        if(!Disparity::contains(disparities, keypointLeft.pt)){
            double value = 0;
            int counter = 0;
            for (unsigned int j = 0; j < matches.size(); j++) {
                if(i!=j){
                    DMatch matchNext = matches.at(j);
                    KeyPoint keypointLeftNext = keypointsLeft.at(matchNext.queryIdx);
                    KeyPoint keypointRightNext = keypointsRight.at(matchNext.trainIdx);
                    double xL = keypointLeft.pt.x - keypointLeftNext.pt.x;
                    double xR = keypointRight.pt.x - keypointRightNext.pt.x;
                    if (xL && xR){
                        double yL = keypointLeft.pt.y - keypointLeftNext.pt.y;
                        double yR = keypointRight.pt.y - keypointRightNext.pt.y;
                        double a = (yR - yL)/xL;
                        double angle = atan(a) * 180 / PI;
                        value += angle;
                        counter++;
                    }
                }
            }
            if (counter > 0){
                value /= counter;
            }
            Disparity* disparity = new Disparity(keypointLeft.pt, value);
            disparities.push_back(disparity);
        }
    }
    double stop = clock();
    double time = 1000*(stop-start)/CLOCKS_PER_SEC;

    sort(disparities.begin(), disparities.end(), Disparity::comparator);
    vector<Disparity*> values(disparities.begin() + (int) disparities.size()*0.1, disparities.end() - (int) disparities.size()*0.1);

    if(values.size()>0){
        data.push_back(new Result::KeyValue("ALL POINTS", matches.size()));
        data.push_back(new Result::KeyValue("CALCULATION POINTS", Statistic::points(values)));
        data.push_back(new Result::KeyValue("AVERAGE", Statistic::average(values)));
        data.push_back(new Result::KeyValue("AVERAGE STANDARD DEVIATION", Statistic::averageSD(values)));
        data.push_back(new Result::KeyValue("AVERAGE ABS", Statistic::averageABS(values)));
        data.push_back(new Result::KeyValue("MEDIAN", Statistic::median(values)));
        data.push_back(new Result::KeyValue("MEDIAN ABSOLUTE DEVIATION",  Statistic::medianAD(values)));
        data.push_back(new Result::KeyValue("CENTYL 70", Statistic::max(values)));
   }
   data.push_back(new Result::KeyValue("TIME", time));


   string distribution = this->getParameter("DISTRIBUTION");
   if(!distribution.compare("TRUE")){
       Mat circles = Visualize::circles(left, values);
       images.push_back(new Result::KeyImage("CIRCLES", circles));
   }

   Mat leftCorrect = correct(left, 0);
   images.push_back(new Result::KeyImage("CORRECT_LEFT", leftCorrect));

   Mat rightCorrect = correct(right, Statistic::median(values));
   images.push_back(new Result::KeyImage(string("CORRECT_RIGHT"), rightCorrect));

   Mat after = Visualize::anaglyph(leftCorrect, rightCorrect);
   images.push_back(new Result::KeyImage(string("ANAGLYPH_AFTER"), after));

   Mat points = Visualize::matches(left, right, correlation);
   images.push_back(new Result::KeyImage(string("MATCHES"), points));

   Result* result = new Result("ROTATION", data, images);
   results.push_back(result);

  return results;
}

Mat RotationDisparityChecker::summary(vector<Result*> results){
    Mat summary(200, 200, CV_8UC3, Scalar(0,0,0));
    return summary;
}
